require 'net/http'
require 'http-cookie'

class ArkService
  attr_accessor :base_url

  def initialize
    @jar = HTTP::CookieJar.new
    @base_url = ENV.fetch('ARK_ENDPOINT', '')
    @namespace = ENV.fetch('ARK_NAMESPACE', '')
    @naan = ENV.fetch('ARK_NAAN', '')
  end

  def authenticate
    username = ENV.fetch('ARK_USERNAME', '')
    password = ENV.fetch('ARK_PASSWORD', '')
    uri = URI(@base_url + 'login')

    request = Net::HTTP::Post.new(uri)
    request.basic_auth username, password
    success, _response_body, response = _do_request(request)

    if success
      response.get_fields('set-cookie')&.each do |val|
        @jar.parse(val, request.uri)
      end
    end
    return success, _response_body
  end

  def session_status
    uri = URI(@base_url + 'check_session')
    request = Net::HTTP::Post.new(uri)
    request['Cookie'] = HTTP::Cookie.cookie_value(@jar.cookies(uri))

    success, _response_body, _response = _do_request(request)
    return success, _response_body
  end

  #----------------------------------------------
  # users
  #----------------------------------------------
  def list_users
    uri = URI(@base_url + 'user')
    request = Net::HTTP::Get.new(uri)
    success, response_body, _response = _do_request(request)
    return response_body unless success
    response_body.fetch('data', [])
  end

  def create_user(name, email, password)
    uri = URI(@base_url + 'user/' + email)
    request = Net::HTTP::Post.new(uri)
    request.set_form_data({
                            'name' => name,
                            'password' => password
                          })
    success, _response_body, _response = _do_request(request)
    return success, _response_body
  end

  def get_user(email)
    uri = URI(@base_url + 'user/' + email)
    request = Net::HTTP::Get.new(uri)
    success, response_body, _response = _do_request(request)
    return response_body unless success
    response_body.fetch('data', [])
  end

  def update_user_details(email, new_name=nil, new_password=nil)
    uri = URI(@base_url + 'user/' + email)
    request = Net::HTTP::Patch.new(uri)
    data = {}
    data['name'] = new_name if new_name
    data['password'] = new_password if new_password
    return false unless data
    data['user'] = email
    request.set_form_data(data)
    success, _response_body, _response = _do_request(request)
    return success, _response_body
  end

  def delete_user(email)
    # Note: ACan only delete logged in user.
    uri = URI(@base_url + 'user/' + email + '/delete')
    request = Net::HTTP::Delete.new(uri)
    success, _response_body, _response = _do_request(request)
    return success, _response_body
  end

  #----------------------------------------------
  # namespaces
  #----------------------------------------------
  def list_namespaces
    uri = URI(@base_url + 'namespace')
    request = Net::HTTP::Get.new(uri)
    success, response_body, _response = _do_request(request)
    return response_body unless success
    response_body.fetch('data', [])
  end

  def create_namespace(namespace, description=nil)
    # Create a new namespace with the current user as admin
    uri = URI(@base_url + 'namespace/' + namespace)
    request = Net::HTTP::Post.new(uri)
    if description
      data = {'desc' => description} if description
      request.set_form_data(data)
    end
    success, _response_body, _response = _do_request(request)
    return success, _response_body
  end

  def get_namespace(namespace)
    uri = URI(@base_url + 'namespace/' + namespace)
    request = Net::HTTP::Get.new(uri)
    success, response_body, _response = _do_request(request)
    return response_body unless success
    response_body.fetch('data', [])
  end

  def delete_namespace(namespace)
    # Note: Only admin user can delete a namespace.
    uri = URI(@base_url + 'namespace/' + namespace)
    request = Net::HTTP::Delete.new(uri)
    success, _response_body, _response = _do_request(request)
    return success, _response_body
  end

  #----------------------------------------------
  # namespace users
  #----------------------------------------------
  def add_namespace_user(namespace, email)
    # Add the specified user to a namespace
    # If admin, add the specified user as an admin to the namespace
    uri = URI(@base_url + 'namespaceuser/' + namespace)
    request = Net::HTTP::Post.new(uri)
    return false unless email
    data = {'email' => email}
    request.set_form_data(data)
    success, _response_body, _response = _do_request(request)
    return success, _response_body
  end

  def set_namespace_user_as_admin(namespace, email)
    # Add the specified user as an admin to the namespace
    uri = URI(@base_url + 'namespaceuser/' + namespace + '/admin')
    request = Net::HTTP::Post.new(uri)
    return false unless email
    data = {'email' => email}
    request.set_form_data(data)
    success, _response_body, _response = _do_request(request)
    return success, _response_body
  end

  def delete_namespace_user(namespace, email)
    # Remove the specified user from the namespace
    uri = URI(@base_url + 'namespaceuser/' + namespace)
    request = Net::HTTP::Delete.new(uri)
    return false unless email
    data = {'email' => email}
    request.set_form_data(data)
    success, _response_body, _response = _do_request(request)
    return success, _response_body
  end

  #----------------------------------------------
  # ARKs
  #----------------------------------------------
  def list_arks
    uri = URI(@base_url + "ark:/#{@naan}/#{@namespace}")
    request = Net::HTTP::Get.new(uri)
    success, response_body, _response = _do_request(request)
    return response_body unless success
    response_body.fetch('data', [])
  end

  def get_ark(ark)
    uri = URI(@base_url + "ark:/#{@naan}/#{@namespace}/#{ark}")
    request = Net::HTTP::Get.new(uri)
    success, response_body, _response = _do_request(request)
    return response_body unless success
    response_body.fetch('data', [])
  end

  def create_ark(ark, work_url, metadata)
    where = "ark:/#{@naan}/#{@namespace}/#{ark}"
    uri = URI(@base_url + where)
    request = Net::HTTP::Post.new(uri)
    data = _prepare_metadata(where, work_url, metadata)
    # request.body = data
    request.set_form_data(data)
    success, _response_body, _response = _do_request(request)
    return success, _response_body, uri.to_s
  end

  def _prepare_metadata(where, work_url, metadata)
    metadata = metadata.with_indifferent_access
    data = {}
    data['_t'] = work_url
    data['how'] = "(:mtype data) dataset"
    data['who'] = metadata[:creator] if metadata.fetch(:creator, nil)
    data['what'] = metadata[:title] if metadata.fetch(:title, nil)
    data['when'] = metadata[:date_created] if metadata.fetch(:date_created, nil)
    data['where'] = where
    data
  end
end


def _do_request(request)
  base_uri = URI(@base_url)
  request_options = { use_ssl: base_uri.scheme == "https" }
  request['Cookie'] = HTTP::Cookie.cookie_value(@jar.cookies(base_uri)) if @jar.cookies(base_uri)
  response = Net::HTTP.start(base_uri.hostname, base_uri.port, request_options) do |http|
    http.request(request)
  end
  success = false
  begin
    response_body = JSON.parse(response.body)
  rescue
    response_body = response.body
  end
  success = true if response_body['error'] == false and response.response.code.to_i.between?(200, 299)
  return success, response_body, response
end

# Usage: to integrate with Hyrax
# a = ArkService.new
# a.authenticate
# work_url = "http://localhost:3000/concern/crc_datasets/tq57nr00k"
# metadata = {'title': 'Experiment A95', 'creator': 'Jones, Ricahrd; Nandakumar, Raja', 'date_created': "2023-01-28T07:51:14Z"}
# ark = "tq57nr00k"
# a.create_ark(ark, work_url, metadata)
