namespace :rdms do
  desc 'Add users to RDMS. usage: rdms:setup_users["users.json", "false"]'
  task :setup_users, [:seedfile, :update_users] => :environment do |task, args|

    seedfile = args.seedfile
    update_user = args.update_users
    if update_user.is_a? String
      if update_user.downcase == 'false'
        update_user = false
      else
        update_user = true
      end
    end
    
    if (File.exists?(seedfile))
      puts("Running task to add / update users from seedfile: #{seedfile}")
    else
      abort("ERROR: missing seedfile for user setup: #{seedfile}")
    end

    seed = JSON.parse(File.read(seedfile))

    seed.fetch('users', []).each_with_index do |parsed_user, index|
      messages = []
      user_count = index + 1
      email = nil
      user = nil
      if parsed_user.fetch('email', nil).present?
        user = User.find_by(email: parsed_user['email'])
        email = parsed_user['email']
      elsif parsed_user.fetch('saml_id', nil).present?
        user = User.find_by(provider: "saml", uid: parsed_user['saml_id'])
        if user.present?
          email = user.email
        else
          email = parsed_user['saml_id']
          email = "#{parsed_user['saml_id']}@saml" unless parsed_user['saml_id'].include?('@')
        end
      elsif parsed_user.fetch('orcid', nil).present?
        orcid = parsed_user['orcid']
        orcid = "https://orcid.org/#{parsed_user['orcid']}" unless orcid.start_with?("https://orcid.org/")
        user = User.find_by(orcid: orcid)
        if user.present?
          email = user.email
        else
          email = "#{parsed_user['orcid']}@orcid"
        end
      end

      if user
        if update_user
          messages << "Info: Updating user #{email}"
        else
          messages << "Warning: User #{email} already exists. Not updating user."
          user = nil
        end
      elsif email.blank?
        messages << "Error: Not creating user #{user_count}. User has no email, saml_id or orcid."
      elsif parsed_user.fetch('saml_id', nil).blank? and
        parsed_user.fetch('orcid', nil).blank? and parsed_user.fetch('password', nil).blank?
        messages << "Error: Not creating user #{email}. User has no password, saml_id or orcid."
      else
        messages << "Info: Creating user #{email}"
        user = User.new
      end

      if user.present?
        user.email = email

        if parsed_user.fetch('password', nil).present?
          user.password = parsed_user['password']
        else
          user.password = SecureRandom.random_bytes(32)
        end

        user.display_name = parsed_user["name"] if parsed_user.fetch('name', nil).present?

        if parsed_user.fetch('saml_id', nil).present?
          user.provider = "saml"
          user.uid = parsed_user['saml_id']
        end

        user.orcid = parsed_user['orcid'] if parsed_user.fetch('orcid', nil).present?

        if user.save
          _success, message = assign_user_to_role(user, parsed_user)
          messages << message
        else
          messages << "Error: #{user.errors.full_messages}"
        end
      end

      if messages.any?
        puts "--- User: #{user_count} ---"
        puts messages
      end
    end
  end

  def assign_user_to_role(user, parsed_user)
    role_name = parsed_user.fetch('role', nil)
    group_id = parsed_user.fetch('group_id', nil)
    return true, '' if role_name.blank?

    role = if %w(admin publication_manager crc_1280_manager crc_1280_member).include?(role_name)
      Role.find_or_create_by(name: role_name)
    elsif role_name == "crc_1280_group_manager" and group_id.present?
      Role.find_or_create_by(name: "crc_1280_#{group_id}_manager")
    elsif role_name == "crc_1280_group_member" and group_id.present?
      Role.find_or_create_by(name: "crc_1280_#{group_id}_member")
    end

    unless role.present?
      if %w(crc_1280_group_manager crc_1280_group_member).include?(role_name) and group_id.blank?
        return false, "Error: #{role_name} needs a group id"
      else
        return false, "Error: #{role_name} is not a valid role"
      end
    end

    role.users << user if role.users.where(id: user.id).blank?

    return true, "Assigned user #{user.email} role #{role.name}"
  end
end

# Example json file
# {
#   "users":
#   [{
#     "email": "example_user@hyrax",
#     "password": "password",
#     "name": "Example User",
#     "saml_id": "urn:oasis:names:tc:SAML:attribute:pairwise-id",
#      "orcid": "orcid id",
#     "role": "crc_1280_group_manager",
#     group_id: "vd66vz88c"
#   }]
# }
